#!/usr/bin/perl -W
#
###############################################################
#   Copyright (C) tridge@samba.org, 2001 (some utility functions)
#   Copyright (C )Stefan Metzmacher <metze@samba.org> 2008
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

use strict;
use warnings;

use Getopt::Long;

use POSIX;

use Time::gmtime;
use Time::Local;

use Data::Dumper;

my $opt_help = 0;
my $opt_verbose = 0;
my $opt_mode = "getdata";
my $opt_name = undef;
my $opt_type = undef;
my @opt_addresses = ();
my $opt_local_owner = undef;
my @opt_remote_wins_servers = ();

my $opt_status_file = undef;
my $opt_log_file = undef;

my $opt_ldbsearch = "ldbsearch";
my $opt_nmblookup = "nmblookup";

my $opt_wins_ldb_url="/var/lib/samba4wins/wins.ldb";
my $opt_expire_warn_interval=0;

my @global_wins_servers = ();


##############################################
# load a list from a file, using : to separate
sub load_list($)
{
	my $fname = shift;
	my @lines = split('\n', `grep -v ^# $fname`);
	return @lines;
}

##############################################
# load a hash from a file, using : to separate
sub load_hash($)
{
	my $fname = shift;
	my @lines = load_list($fname);
	my %ret;
	for my $l (@lines) {
		if ($l =~ /^([\w\-]*)\s*:\s*(.*)$/) {
			$ret{$1} = $2;
		}
	}
	return %ret;
}

#####################################################################
# check if a string is in an array
sub InArray($$)
{
	my ($s, $a) = @_;
	for my $v (@{$a}) {
		return 1 if ($v eq $s);
	}
	return 0;
}

#####################################################################
# flatten an array of arrays into a single array
sub FlattenArray($)
{
	my $a = shift;
	my @b;
	for my $d (@{$a}) {
		push(@b, $_) foreach (@{$d});
	}
	return \@b;
}

#####################################################################
# flatten an array of hashes into a single hash
sub FlattenHash($)
{
	my $a = shift;
	my %b;
	for my $d (@{$a}) {
		for my $k (keys %{$d}) {
			$b{$k} = $d->{$k};
		}
	}
	return \%b;
}

#####################################################################
# return the modification time of a file
sub FileModtime($)
{
	my($filename) = shift;
	return (stat($filename))[9];
}

#####################################################################
# read a file into a string
sub FileLoad($)
{
	my($filename) = shift;
	local(*INPUTFILE);
	open(INPUTFILE, $filename) || return "";
	my($saved_delim) = $/;
	undef $/;
	my($data) = <INPUTFILE>;
	close(INPUTFILE);
	$/ = $saved_delim;
	return $data;
}

#####################################################################
# write a string into a file
sub FileSave($$)
{
	my($filename) = shift;
	my($v) = shift;
	local(*FILE);
	open(FILE, ">$filename") || die "can't open $filename";
	print FILE $v;
	close(FILE);
}

#####################################################################
# return a filename with a changed extension
sub ChangeExtension($$)
{
	my($fname,$ext) = @_;
	return "$1.$ext" if ($fname =~ /^(.*)\.(.*?)$/);
	return "$fname.$ext";
}

#####################################################################
# save a data structure into a file
sub SaveStructure($$)
{
	my($filename) = shift;
	my($v) = shift;
	FileSave($filename, Dumper($v));
}

#####################################################################
# load a data structure from a file (as saved with SaveStructure)
sub LoadStructure($)
{
	my ($fname) = @_;
	# it's important to prefix 'my ' otherwise
	# 'use strict' will cause problems
	my $data = "my ".FileLoad($fname);
	return eval $data;
}

sub pull_ldb_time($)
{
	my ($ldbtime) = @_;

	return 0 unless defined($ldbtime);

	my $year = substr($ldbtime,0,4);
	my $fyear = $year - 1900;
	my $mon  = substr($ldbtime,4,2);
	my $fmon  = $mon - 1;
	my $day  = substr($ldbtime,6,2);
	my $hour = substr($ldbtime,8,2);
	my $min  = substr($ldbtime,10,2);
	my $sec  = substr($ldbtime,12,2);

	my $unixtime = timegm($sec, $min, $hour, $day, $fmon, $fyear);

	return $unixtime;
}

sub push_ldb_time($)
{
	my ($t) = @_;

	my $gm = gmtime($t);

	my $year = $gm->year();
	my $fyear = $year + 1900;
	my $mon  = $gm->mon();
	my $fmon  = $mon + 1;
	my $day  = $gm->mday();
	my $hour = $gm->hour();
	my $min  = $gm->min();
	my $sec  = $gm->sec();

	my $ret = sprintf("%04d%02d%02d%02d%02d%02d.0Z",
			  $fyear,$fmon,$day,
			  $hour,$min,$sec);

	return $ret;
}

sub parse_netstat_line($)
{
	my ($line) = @_;
	my $l = undef;

	$line =~ s/\t/ /g;
	$line =~ s/ +/ /g;

	my @chunks = split(' ',$line);

	#$l->{line} = $line;

	$l->{proto} = $chunks[0];

	my $socket = $chunks[3];

	my @s = split(':',$socket);
	$l->{port} = pop(@s);
	$l->{addr} = join(':',@s);

	my $process = "";
	if ($l->{proto} eq "tcp") {
		splice(@chunks, 0, 6);
		$process = join(' ', @chunks);
	} elsif ($l->{proto} eq "udp") {
		splice(@chunks, 0, 5);
		$process = join(' ', @chunks);
	}

	my @p = split('/',$process);
	$l->{pid} = shift(@p);
	$l->{comm} = join('/',@p);

	return $l;
}

sub dump_netstat_listen($)
{
	my ($port) = @_;
	my $stat = undef;

	my $netstat="netstat -l -p -n -t -u";

	#print "${netstat}";
	@{$stat->{lines}} = `${netstat}`;

	my @plines = grep(/:$port /, @{$stat->{lines}});
	foreach my $l (@plines) {
		push(@{$stat->{addrs}}, parse_netstat_line($l));
	}

	delete $stat->{lines};
	return $stat;
}

sub parse_ps_o_line($$)
{
	my ($line,$comm) = @_;
	my $l = undef;

	my @c = split($comm,$line);

	my $l1 = shift(@c);
	my $l2 = shift(@c);
	my $l3 = join($comm, @c);

	$l1 =~ s/\t/ /g;
	$l2 =~ s/ +/ /g;
	$l3 =~ s/ +/ /g;

	my @c1 = split(' ',$l1);
	my @c2 = split(' ',$l2);
	my @c3 = split(' ',$l3);

	#$l->{l1} = $l1;
	#$l->{l2} = $l2;
	#$l->{l3} = $l3;
	#@{$l->{c1}} = @c1;
	#@{$l->{c2}} = @c2;
	#@{$l->{c3}} = @c3;

#ps -C $comm -o user,pid,ppid,stat,vsize,rssize,size,cputime,%cpu,%mem,comm,start,comm,cmd

	$l->{user}	= $c1[0];
	$l->{pid}	= $c1[1];
	$l->{ppid}	= $c1[2];
	$l->{state}	= $c1[3];
	$l->{vsize}	= $c1[4];
	$l->{rssize}	= $c1[5];
	$l->{size}	= $c1[6];
	$l->{cputime}	= $c1[7];
	$l->{cpu}	= $c1[8];
	$l->{mem}	= $c1[9];

	$l->{comm}	= $comm;

	$l->{start}	= join(' ', @c2);

	$l->{cmd}	= join(' ', @c3);

	return $l;
}

sub dump_processes($)
{
	my ($comm) = @_;
	my $ps = undef;

	my $ps_o="ps -C $comm -o user,pid,ppid,stat,vsize,rssize,size,cputime,%cpu,%mem,comm,start,comm,cmd";

	#print "${ps_o}";
	@{$ps->{lines}} = `${ps_o}`;

	my @lines = grep(/$comm/, @{$ps->{lines}});
	foreach my $l (@lines) {
		push(@{$ps->{pids}}, parse_ps_o_line($l, $comm));
	}

	delete $ps->{lines};
	return $ps;
}

sub parse_ldb_addr_line($$)
{
	my ($rec,$line) = @_;
	my $addr = undef;

	$line =~ s/^address: //;

	my @chunks = split(/;/,$line);

	$addr->{addr} = $chunks[0];
	$addr->{owner} = $chunks[1];
	$addr->{owner} =~ s/winsOwner://;
	$addr->{expire_string} = $chunks[2];
	$addr->{expire_string} =~ s/expireTime://;
	$addr->{expire_secs} = pull_ldb_time($addr->{expire_string});
	#$addr->{expire_string2} = push_ldb_time($addr->{expire_secs});

	if ($addr->{expire_secs} < $rec->{now_secs}) {
		$addr->{is_expired} = 1;
	} else {
		$addr->{is_expired} = 0;
	}

	if (($addr->{expire_secs} - $opt_expire_warn_interval) < $rec->{now_secs}) {
		$addr->{will_expired} = 1;
	} else {
		$addr->{will_expired} = 0;
	}

	return $addr;
}

sub get_ldb_values($$)
{
	my ($lines, $attr) = @_;

	return undef unless defined($lines);

	my @alines = grep(/^$attr: /, @{$lines});

	foreach my $l (@alines) {
		chomp $l;
		$l =~ s/^$attr: //;
	}

	return \@alines;
}

sub get_ldb_value($$)
{
	my ($lines, $attr) = @_;

	my $vals = get_ldb_values($lines, $attr);

	return undef unless defined($vals);

	return @{$vals}[0];
}

sub load_ldb_name($$$)
{
	my ($rec,$name,$type) = @_;
	my $ldb = undef;

	$ldb->{dn} = "name=${name},type=0x${type}";

	my $ldbsearch = "${opt_ldbsearch} --url='${opt_wins_ldb_url}' --basedn='$ldb->{dn}' --scope='base' -d 0";

	#print "${ldbsearch}";
	@{$ldb->{lines}} = `${ldbsearch}`;

	return undef unless defined($ldb->{lines});

	my @dn = grep(/^dn: /, @{$ldb->{lines}});

	return undef unless @dn == 1;

	my @addrlines = grep(/^address: /, @{$ldb->{lines}});

	$ldb->{num_addrs} = 0;
	foreach my $l (@addrlines) {
		my $a = parse_ldb_addr_line($rec, $l);
		push(@{$ldb->{addrs}}, $a);
		$ldb->{num_addrs}++;
	}

	$ldb->{owner} = get_ldb_value($ldb->{lines}, "winsOwner");
	if ($ldb->{owner} eq $opt_local_owner) {
		$ldb->{is_local} = 1;
	} else {
		$ldb->{is_local} = 0;
	}

	$ldb->{expire_string} = get_ldb_value($ldb->{lines}, "expireTime");
	$ldb->{expire_secs} = pull_ldb_time($ldb->{expire_string});
	#$ldb->{expire_string2} = push_ldb_time($ldb->{expire_secs});

	if ($ldb->{expire_secs} < $rec->{now_secs}) {
		$ldb->{is_expired} = 1;
	} else {
		$ldb->{is_expired} = 0;
	}

	if (($ldb->{expire_secs} - $opt_expire_warn_interval) < $rec->{now_secs}) {
		$ldb->{will_expire} = 1;
	} else {
		$ldb->{will_expire} = 0;
	}

	$ldb->{is_static} = get_ldb_value($ldb->{lines}, "isStatic");

	$ldb->{state} = get_ldb_value($ldb->{lines}, "recordState");

	delete $ldb->{lines};
	return $ldb;
}

sub parse_nmblookup_addr_line($)
{
	my ($line) = @_;

	my $addr = $line;
	chomp $addr;
	$addr =~ s/([^ ]+) .*/$1/;

	return $addr;
}

sub nmblookup_name_wins($$$)
{
	my ($name,$type,$server) = @_;
	my $lookup = undef;

	$lookup->{name} = "${name}#${type}";
	$lookup->{server} = $server;

	my $nmblookup = "${opt_nmblookup} --configfile /dev/null -d 0 -R -U $lookup->{server} $lookup->{name}";

	#print "${nmblookup}\n";
	@{$lookup->{lines}} = `${nmblookup}`;

	my @addrlines = grep(/ $name</, @{$lookup->{lines}});

	$lookup->{num_addrs} = 0;
	foreach my $l (@addrlines) {
		push(@{$lookup->{addrs}}, parse_nmblookup_addr_line($l));
		$lookup->{num_addrs}++;
	}

	delete $lookup->{lines};
	return $lookup;
}

sub nmblookup_name_unicast($$$)
{
	my ($name,$type,$ip) = @_;
	my $lookup = undef;

	$lookup->{name} = "${name}#${type}";
	$lookup->{ip} = $ip;

	my $nmblookup = "${opt_nmblookup} --configfile /dev/null -d 0 -U $lookup->{ip} $lookup->{name}";

	#print "${nmblookup}\n";
	@{$lookup->{lines}} = `${nmblookup}`;

	my @addrlines = grep(/ $name</, @{$lookup->{lines}});

	$lookup->{num_addrs} = 0;
	foreach my $l (@addrlines) {
		my $a = parse_nmblookup_addr_line($l);
		# samba3 has a bug and returns 0.0.0.0 as an extra address
		# for #1C names
		next if ($a eq "0.0.0.0");
		push(@{$lookup->{addrs}}, $a);
		$lookup->{num_addrs}++;
	}

	delete $lookup->{lines};
	return $lookup;
}

sub parse_nmblookup_name_line($)
{
	my ($line) = @_;
	my $name = undef;

	chomp $line;
	$line =~ s/^\t(.*)/$1/;

	$name->{name} = substr($line, 0, 15);
	$name->{name} =~ s/\ *$//;
	$name->{type} = uc(substr($line, 17, 2));

	my $rest = substr($line, 23, -1);

	if ($rest =~ /\<GROUP\>/) {
		$name->{GROUP} = 1;
	} else {
		$name->{GROUP} = 0;
	}

	if ($rest =~ /\<ACTIVE\>/) {
		$name->{ACTIVE} = 1;
	} else {
		$name->{ACTIVE} = 0;
	}

	return $name;
}

sub nmblookup_name_stat($)
{
	my ($ip) = @_;
	my $lookup = undef;

	$lookup->{ip} = $ip;

	my $nmblookup = "${opt_nmblookup} --configfile /dev/null -d 0 -A $lookup->{ip}";

	#print "${nmblookup}\n";
	@{$lookup->{lines}} = `${nmblookup}`;

	my @namelines = grep(/ </, @{$lookup->{lines}});

	$lookup->{num_names} = 0;
	foreach my $l (@namelines) {
		push(@{$lookup->{names}}, parse_nmblookup_name_line($l));
		$lookup->{num_names}++;
	}

	delete $lookup->{lines};
	return $lookup;
}

sub rec_report($$$$)
{
	my ($rec, $logfd, $verbose, $oneline) = @_;
	my $ret = 0;
	my $header = "";
	my $status = "OK";
	my $name = "";
	my $info = "";
	my $warn = "";
	my $err = "";

	if (not defined($rec)) {
		print $logfd "UNKNOWN: No valid Data available in [$opt_status_file]\n";
		return 3;
	}

	$header .= "$rec->{now_string}: Local-Owner[$rec->{local_owner}]";

	$name .= "Name[$rec->{name}] Type[$rec->{type}]";

	$info .= "Expected addresses[$rec->{expected}->{num_addrs}]:\n";
	foreach my $addr (@{$rec->{expected}->{addrs}}) {
		$info .= "\t$addr\n";
	}

	if (defined($rec->{warnings})) {
		$ret = 1;
		$status = "WARN";
		$warn .= "Warnings[$rec->{num_warnings}]:\n";
		foreach my $w (@{$rec->{warnings}}) {
			my $location = "\tWarning";
			$location = "Warning[$w->{func}():$w->{line}] " if $verbose;
			$warn .= "${location}: $w->{msg}\n";
		}
	}

	if (defined($rec->{errors})) {
		$ret = 2;
		$status = "ERROR";
		$err .= "Errors[$rec->{num_errors}]:\n";
		foreach my $e (@{$rec->{errors}}) {
			my $location = "\tError";
			$location = "Error[$e->{func}():$e->{line}]" if $verbose;
			$err .= "${location}: $e->{msg}\n";
		}
	}

	my $out = "";

	$out .= "$header\n" unless $oneline;

	$out .= "$status: $name";
	$out .= ": Checked[$rec->{now_string}]"  if $oneline;
	$out .= "\n";

	$out .= $info if ($verbose or ($ret != 0 and not $oneline));
	$out .= $warn unless $oneline;
	$out .= $err unless $oneline;

	$out .= Data::Dumper::Dumper($rec) if $verbose;

	print $logfd $out;

	return $ret;
}

sub rec_warn($$)
{
	my ($rec, $msg) = @_;
	my $warn = undef;

	my @c0 = caller(0);
	my @c1 = caller(1);

	$warn->{func}	= $c1[3];
	$warn->{line}	= $c0[2];
	#$warn->{c0}	= \@c0;
	#$warn->{c1}	= \@c1;
	$warn->{msg}	= $msg;

	$rec->{num_warnings}++;
	push(@{$rec->{warnings}}, $warn);
}

sub rec_error($$)
{
	my ($rec, $msg) = @_;
	my $error = undef;

	my @c0 = caller(0);
	my @c1 = caller(1);

	$error->{func}	= $c1[3];
	$error->{line}	= $c0[2];
	#$error->{c0}	= \@c0;
	#$error->{c1}	= \@c1;
	$error->{msg}	= $msg;

	$rec->{num_errors}++;
	push(@{$rec->{errors}}, $error);
}

sub create_addrs($)
{
	my ($rec) = @_;

	$rec->{addrs} = undef;

	my $src = "expected";
	$rec->{sources}->{$src} = $rec->{expected}->{num_addrs};
	foreach my $addr (@{$rec->{expected}->{addrs}}) {
		$rec->{addrs}->{$addr}->{$src}++;
	}

	$src = "ldb";
	if (defined($rec->{ldb})) {
		$rec->{sources}->{$src} = $rec->{ldb}->{num_addrs};
		foreach my $a (@{$rec->{ldb}->{addrs}}) {
			$rec->{addrs}->{$a->{addr}}->{$src}++;
		}
	} else {
		$rec->{sources}->{$src} = 0;
	}

	foreach my $l (@{$rec->{wins}}) {
		$src = "wins:$l->{server}";
		$rec->{sources}->{$src} = $l->{num_addrs};
		foreach my $addr (@{$l->{addrs}}) {
			$rec->{addrs}->{$addr}->{$src}++;
		}
	}

	foreach my $l (@{$rec->{unicast}}) {
		$src = "unicast:$l->{ip}";
		$rec->{sources}->{$src} = $l->{num_addrs};
		foreach my $addr (@{$l->{addrs}}) {
			$rec->{addrs}->{$addr}->{$src}++;
		}
	}
}

sub check_addrs($)
{
	my ($rec) = @_;

	foreach my $src (keys %{$rec->{sources}}) {
		my $expected = $rec->{sources}->{expected};
		my $count = $rec->{sources}->{$src};
		next unless $count != $expected;

		if (not $rec->{is_sgroup}) {
			rec_error($rec, "$src: num_addrs[$count] expected[$expected]");
			next;
		}
		if (not ($src =~ /^unicast:/)) {
			rec_error($rec, "$src: num_addrs[$count] expected[$expected]");
			next;
		}
		if ($count == 0) {
			rec_error($rec, "$src: num_addrs[$count] expected[$expected]");
			next;
		}
		if ($count > $expected) {
			rec_error($rec, "$src: num_addrs[$count] expected[$expected]");
			next;
		}

		#rec_warn($rec, "$src: num_addrs[$count] expected[$expected]");
	}

	foreach my $addr (keys %{$rec->{addrs}}) {
		foreach my $src (keys %{$rec->{addrs}->{$addr}}) {
			my $expected = $rec->{addrs}->{$addr}->{expected};
			$expected = 0 unless defined($expected);
			my $count = $rec->{addrs}->{$addr}->{$src};
			next unless $count != $expected;

			if ($count > $expected) {
				rec_error($rec, "$src: $addr\[$count] expected[$expected]");
				# this address is not expected see if it really holds the name
				my $stat = nmblookup_name_stat($addr);
				$stat->{unexpected} = $src;
				push(@{$rec->{stats}}, $stat);
				next;
			}
			if (not $rec->{is_sgroup}) {
				rec_error($rec, "$src: $addr\[$count] expected[$expected]");
				next;
			}
			if (not ($src =~ /^unicast:/)) {
				rec_error($rec, "$src: $addr\[$count] expected[$expected]");
				next;
			}

			#rec_warn($rec, "$src: $addr\[$count] expected[$expected]");
		}
	}

	foreach my $l (@{$rec->{stats}}) {
		my $found = 0;

		foreach my $n (@{$l->{names}}) {
			next unless ($n->{name} eq $rec->{name});
			next unless ($n->{type} eq $rec->{type});

			$found = 1;

			if (defined($l->{unexpected}) and $n->{ACTIVE}) {
				rec_error($rec, "stat:$l->{ip}: name is active ".
					  "and reported by[$l->{unexpected}] but unexpected");
				last;
			}
			if (not $n->{ACTIVE}) {
				rec_error($rec, "stat:$l->{ip}: name not active");
			}
			if (not $rec->{is_sgroup} and $n->{GROUP}) {
				rec_error($rec, "stat:$l->{ip}: name is a group name");
			}
			if ($rec->{is_sgroup} and not $n->{GROUP}) {
				rec_error($rec, "stat:$l->{ip}: name is not a group name");
			}

			last;
		}
		if (not defined($l->{unexpected}) and $found == 0) {
			rec_error($rec, "stat:$l->{ip}: name not returned");
			next;
		}
	}
}

sub time_interval_left($$)
{
	my ($high, $low) = @_;
	my $res = undef;
	my $diff = $high - $low;

	if ($diff < 0) {
		$res->{sign} = -1;
	} else {
		$res->{sign} = 1;
	}
	$res->{diff} = $diff;

	$diff *= $res->{sign};

	$res->{days} = int($diff / (24 * 60 * 60));
	$diff = $diff % (24 * 60 * 60);

	$res->{hours} = int($diff / (60 * 60));
	$diff = $diff % (60 * 60);

	$res->{mins} = int($diff / (60));
	$diff = $diff % (60);

	$res->{secs} = $diff;

	$res->{str} = "";
	$res->{str} .= "in " if ($res->{sign} > 0);

	$res->{str} .= "$res->{days}d " if ($res->{days} > 0);

	$res->{str} .= "$res->{hours}h " if ($res->{hours} > 0);

	$res->{str} .= "$res->{mins}m " if ($res->{mins} > 0);

	$res->{str} .= "$res->{secs}s";

	$res->{str} .= " ago" if ($res->{sign} < 0);

	return $res;
}

sub check_expire($)
{
	my ($rec) = @_;

	return unless defined($rec->{ldb});

	if ($rec->{ldb}->{state} != 0) {
		rec_error($rec, "ldb: state[$rec->{ldb}->{state}]");
	}

	if ($rec->{ldb}->{is_local} and $rec->{ldb}->{is_expired}) {
		my $addrs = "";
		foreach my $a (@{$rec->{ldb}->{addrs}}) {
			$addrs .= "\n\t$a->{addr} owner[$a->{owner}] expire[$a->{expire_string}]";
			next;
		}
		rec_error($rec, "[$rec->{now_string}] expired at [$rec->{ldb}->{expire_string}]".
			  $addrs);
		return;
	}

	if ($rec->{ldb}->{is_local} and $rec->{ldb}->{will_expire}) {
		my $addrs = "";
		foreach my $a (@{$rec->{ldb}->{addrs}}) {
			$addrs .= "\n\t$a->{addr} owner[$a->{owner}] expire[$a->{expire_string}]";
			next;
		}
		my $diff = time_interval_left($rec->{ldb}->{expire_secs}, $rec->{now_secs});
		rec_warn($rec, "[$rec->{now_string}] record will expire $diff->{str} at [$rec->{ldb}->{expire_string}]".
			 $addrs);
		return;
	}

	foreach my $a (@{$rec->{ldb}->{addrs}}) {
		next unless ($a->{is_local});

		if ($a->{is_expired}) {
			rec_error($rec, "[$rec->{now_string}]: $a->{addr} ".
				  "expired at [$a->{expire_string}] ".
				  "rec[$rec->{ldb}->{expire_string}]");
			next;
		}

		if ($a->{will_expire}) {
			my $diff = time_interval_left($a->{expire_secs}, $rec->{now_secs});
			rec_warn($rec, "[$rec->{now_string}]: $a->{addr} ".
				 "will expire $diff->{str} at [$a->{expire_string}] ".
				 "rec[$rec->{ldb}->{expire_string}]");
			next;
		}
	}
}

sub check_netstat($$$)
{
	my ($rec,$proto,$port) = @_;

	if (not defined($rec->{netstat})) {
		rec_error($rec, "not listen on any ports");
		return;
	}

	if (not defined($rec->{netstat}->{$port})) {
		rec_error($rec, "not listen on port $port");
		return;
	}

	my $addrs = undef;
	my $unique = 0;
	my $wildcard = 0;
	my $count = 0;
	foreach my $a (@{$rec->{netstat}->{$port}->{addrs}}) {
		next if $a->{proto} ne $proto;
		$count++;

		push(@{$addrs->{$a->{addr}}}, $a);

		my $is_wildcard = 0;
		$is_wildcard = 1 if (($a->{addr} eq "0.0.0.0") or ($a->{addr} eq "::"));

		$unique++ unless $is_wildcard;
		$wildcard++ if $is_wildcard;
	}

	if ($count == 0) {
		rec_error($rec, "0 addrs listen on $proto port $port");
		return;
	}

	if ($wildcard > 0 and $unique > 0) {
		my $err = "$proto port $port some wildcard and unique addresses";
		foreach my $a (@{$rec->{netstat}->{$port}->{addrs}}) {
			next if $a->{proto} ne $proto;
			$err .= "\n\taddr[$a->{addr}] pid[$a->{pid}] comm[$a->{comm}]";
		}
		rec_error($rec, $err);
	}

	foreach my $k (keys %{$addrs}) {
		my $last = 0;
		my $pids = 0;
		foreach my $a (@{$addrs->{$k}}) {
			next if ($last == $a->{pid});

			$last = $a->{pid};
			$pids++;
		}

		last if ($pids == 1);

		my $err = "$proto addr $k port $port multiple processes[$pids]";
		foreach my $a (@{$addrs->{$k}}) {
			$err .= "\n\taddr[$a->{addr}] pid[$a->{pid}] comm[$a->{comm}]";
		}
		rec_error($rec, $err);
	}
}

sub check_processes($)
{
	my ($rec) = @_;

	if (not defined($rec->{processes})) {
		rec_error($rec, "no processes started");
		return;
	}

	if (not defined($rec->{processes}->{smbd4wins})) {
		rec_error($rec, "no smbd4wins process started");
		return;
	}

	my $count = 0;
	foreach (@{$rec->{processes}->{smbd4wins}->{pids}}) {
		$count++;
	}

	if ($count == 0) {
		rec_error($rec, "0 smbd4wins processes started");
		return;
	}
}

sub check_name($$$)
{
	my ($name, $type, $addrs) = @_;
	my $rec = undef;

	$rec->{name} = $name;

	$rec->{type} = uc($type);

	if ($rec->{type} eq "1C") {
		$rec->{is_sgroup} = 1;
	} else {
		$rec->{is_sgroup} = 0;
	}

	$rec->{given}->{addrs} = $addrs;

	$rec->{given}->{num_addrs} = 0;
	foreach my $addr (@{$rec->{given}->{addrs}}) {
		$rec->{given}->{num_addrs}++;
	}

	$rec->{now_secs} = time();
	$rec->{now_string} = push_ldb_time($rec->{now_secs});

	$rec->{local_owner} = $opt_local_owner;

	$rec->{ldb} = load_ldb_name($rec, $rec->{name},$rec->{type});

	if ($rec->{given}->{num_addrs} > 0) {
		$rec->{expected}->{num_addrs} = $rec->{given}->{num_addrs};
		@{$rec->{expected}->{addrs}} = ();
		foreach my $a (@{$rec->{given}->{addrs}}) {
			push(@{$rec->{expected}->{addrs}}, $a);
		}
	} elsif (defined($rec->{ldb})) {
		$rec->{expected}->{num_addrs} = $rec->{ldb}->{num_addrs};
		my $rec_expired = 0;
		$rec_expired = 1 if ($rec->{ldb}->{is_local} and $rec->{ldb}->{is_expired});
		@{$rec->{expected}->{addrs}} = ();
		foreach my $a (@{$rec->{ldb}->{addrs}}) {
			my $a_expired = 0;
			$a_expired = 1 if ($a->{is_local} and $a->{is_expired});
			$a_expired = 1 if ($a->{is_local} and $rec_expired);
			if ($a_expired) {
				$rec->{expected}->{num_addrs}--;
				rec_warn($rec, "expected from ldb: $a->{addr} expired".
					 "rec[$rec->{ldb}->{expire_string}] addr[$a->{expire_string}]");
				next;
			}
			push(@{$rec->{expected}->{addrs}}, $a->{addr});
		}
		if ($rec->{ldb}->{state} != 0) {
			$rec->{expected}->{num_addrs} = 0;
			@{$rec->{expected}->{addrs}} = ();
			rec_error($rec, "expected from ldb: state[$rec->{ldb}->{state}]");
		} elsif ($rec->{expected}->{num_addrs} == 0) {
			rec_error($rec, "expected from ldb: no addresses valid");
		}
	} else {
		$rec->{expected}->{num_addrs} = 0;
		@{$rec->{expected}->{addrs}} = ();
		rec_error($rec, "expected from ldb: no addresses record was not found in ldb");
	}

	foreach my $server (@global_wins_servers) {
		push(@{$rec->{wins}}, nmblookup_name_wins($rec->{name}, $rec->{type}, $server));
	}

	foreach my $a (@{$rec->{expected}->{addrs}}) {
		push(@{$rec->{unicast}}, nmblookup_name_unicast($rec->{name}, $rec->{type}, $a));
		push(@{$rec->{stats}}, nmblookup_name_stat($a));
	}

	$rec->{netstat}->{137} = dump_netstat_listen(137);
	$rec->{netstat}->{42} = dump_netstat_listen(42);
	$rec->{processes}->{smbd4wins} = dump_processes('smbd4wins');

	check_expire($rec);
	create_addrs($rec);
	check_addrs($rec);
	check_netstat($rec,"udp",137);
	check_netstat($rec,"tcp",42);
	check_processes($rec);

	return $rec;
}

sub store_status($$)
{
	my ($rec, $status_file) = @_;

	return unless defined($status_file);

	$rec->{STATUS_FILE_MAGIC} = "StatusFileMagic";

	SaveStructure($status_file, $rec);
}

sub read_status($)
{
	my ($status_file) = @_;
	my $rec = undef;

	$rec = LoadStructure($status_file);

	return undef unless defined($rec);

	return undef unless defined($rec->{STATUS_FILE_MAGIC});

	if ($rec->{STATUS_FILE_MAGIC} eq "StatusFileMagic") {
		delete $rec->{STATUS_FILE_MAGIC};
		return $rec;
	}

	return undef;
}

sub InvalidArgs($)
{
	my ($msg) = @_;

	print "Invalid arguments:\n";
	print "${msg}\n";
	print "try --help\n";
	exit(1);
}

sub ShowHelp()
{
	print "samba4wins-check-name
Copyright (C) Stefan Metzmacher <metze\@samba.org>
Released under the GNU GPL

Usage: samba4wins-check-name.pl [OPTIONS]

Generic options:
 --help|-h|-?			this help page

 --verbose|-v			verbose output

 --mode|-m			The operation mode (default:getdata):

				getdata: collect data for a specified name
				         and analyze it. The results will
					 be printed to STDOUT or the log-file
					 and optinaly to a status-file.

				getstatus: return the status from a status-file
					   that was created by a former run
					   in 'getdata' mode.

 --name|-n			name to lookup: e.g. FOO.
				Only valid in getdata mode.

 --type|-t			name type to lookup: e.g. 1C
				Only valid in getdata mode.

 --address|-a			list of expected addresses
				Only valid in getdata mode and
				allowed multiple times.

 --local-owner|-l		the ip address of the local samba4wins
				Only valid in getdata mode.

 --remote-wins-server|-r	list of remote wins-servers to ask
				This option is optional and only valid in getdata mode
				and allowed multiple times.

 --expire-warn-interval|-w	If an address or record will expire in less than
				this interval (in seconds) a warning is generated.
				This option is optional and the default is 0.
				Only valid in getdata mode.

 --log-file			The log-file that gets the results appended.
				The default is to not use a log-file and
				print the results to STDOUT.
				Only valid in getdata mode.

 --status-file			In status-file that gets created in 'getdata' mode
				and will be used in 'getstatus' mode.
				The default is to not maintain a status-file.
				The option is optional but valid in getdata mode
				and required in getstatus mode.

 --wins-ldb-url			The wins.ldb path.
				This option is optional and the default is '/var/lib/samba4wins/wins.ldb'.

 --ldbsearch-cmd		The command line for ldbsearch.
				This option is optional and the default is 'ldbsearch'.

 --nmblookup-cmd		The command line for nmblookup.
				This option is optional and the default is 'nmblookup'.
";
	exit(0);
}

Getopt::Long::Configure('no_auto_abbrev');

my $result = GetOptions (
	'help|h|?' => \$opt_help,
	'verbose|v' => \$opt_verbose,
	'mode|m=s' => \$opt_mode,
	'name|n=s' => \$opt_name,
	'type|t=s' => \$opt_type,
	'address|a=s@' => \@opt_addresses,
	'local-owner|l=s' => \$opt_local_owner,
	'remote-wins-servers|r=s@' => \@opt_remote_wins_servers,
	'wins-ldb-url=s' => \$opt_wins_ldb_url,
	'expire-warn-interval|w=i' => \$opt_expire_warn_interval,
	'status-file|s=s' => \$opt_status_file,
	'log-file=s' => \$opt_log_file,
	'ldbsearch-cmd=s' => \$opt_ldbsearch,
	'nmblookup-cmd=s' => \$opt_nmblookup,
);

exit(1) if (not $result);

ShowHelp() if ($opt_help);

if ($opt_mode eq "getdata") {
	InvalidArgs("--name required for getdata mode") unless defined($opt_name);
	InvalidArgs("--type required for getdata mode") unless defined($opt_type);
	InvalidArgs("--local-owner required for getdata mode") unless defined($opt_local_owner);

	push(@global_wins_servers, $opt_local_owner);
	push(@global_wins_servers, @opt_remote_wins_servers);

	my $logfd = undef;
	if (defined($opt_log_file)) {
		open($logfd, ">>$opt_log_file") or
			die("Failed to open[$opt_log_file]");
	} else {
		open($logfd, ">-") or
			die("Failed to open STDOUT");;
	}
	my $rec = check_name($opt_name, $opt_type, \@opt_addresses);
	my $ret = rec_report($rec, $logfd, $opt_verbose, 0);
	store_status($rec, $opt_status_file);
	exit($ret);
} elsif ($opt_mode eq "getstatus") {
	InvalidArgs("--status-file required for getstatus mode") unless defined($opt_status_file);
	InvalidArgs("--name not allowed for getstatus mode") if defined($opt_name);
	InvalidArgs("--type not allowed for getstatus mode") if defined($opt_type);
	InvalidArgs("--address not allowed for getstatus mode") if @opt_addresses;
	InvalidArgs("--local-owner not allowed for getstatus mode") if defined($opt_local_owner);
	InvalidArgs("--remote-wins-server not allowed for getstatus mode") if @opt_remote_wins_servers;
	InvalidArgs("--expire-warn-interval not allowed for getstatus mode") if $opt_expire_warn_interval;
	InvalidArgs("--log-file not allowed for getstatus mode") if defined($opt_log_file);

	my $logfd = undef;
	open($logfd, ">-") or
		die("Failed to open STDOUT");;
	my $rec = read_status($opt_status_file);
	my $oneline = 1;
	$oneline = 0 if $opt_verbose;
	my $ret = rec_report($rec, $logfd, $opt_verbose, $oneline);
	exit($ret);
}

ShowHelp();
